/*
Firmware for a bristlebot having a mouse-sensor.
(C) 2009 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#define F_CPU 1000000
#include <avr/io.h>
#include <util/delay.h>

//Connections of the ATTiny13
#define MOTOR_L PB0
#define MOTOR_R PB1
#define ADNS_SDIO PB3
#define ADNS_SCLK PB4
#define ADNS_NCS PB2

#define ADNSWR 0x80

//Mouse sensor comms
//This routine writes a byte to the ADNS5020...
void adnsWriteByte(unsigned char data) {
	unsigned char x;
	PORTB&=~(1<<ADNS_NCS);
	for (x=0; x<8; x++) {
		if (data&0x80) {
			PORTB|=(1<<ADNS_SDIO);
		} else {
			PORTB&=~(1<<ADNS_SDIO);
		}
		PORTB&=~(1<<ADNS_SCLK);
		data<<=1;
		PORTB|=(1<<ADNS_SCLK);
	}
}

//and this one reads a byte from the ADNS5020.
char adnsReadByte(void) {
	char x, ret;
	ret=0;
	DDRB&=~(1<<ADNS_SDIO);
	for (x=0; x<8; x++) {
		PORTB&=~(1<<ADNS_SCLK);
		ret<<=1;
		PORTB|=(1<<ADNS_SCLK);
		if (PINB&(1<<ADNS_SDIO)) ret|=1;
	}
	DDRB|=(1<<ADNS_SDIO);
	return ret;
}

//This routine ends communication with the ADNS5020 by raising
//the CS-line. Invoking adnsWriteByte after this automatically
//lowers it again.
void adnsEnd(void) {
	PORTB|=(1<<ADNS_NCS);
}


//Main routine
int main(void) {
	int x=0, y=0, count=0;
	char dx,dy;
	char isTurning=1; //Start with a turn because that recalibrates the darkness detection
	unsigned char sum, sumCal=0;
	short shut;
	unsigned char squal;
	TCCR0A=0xa3; //fast pwm
	TCCR0B=1; //no prescaling
	
	DDRB=(1<<MOTOR_L)|(1<<MOTOR_R)|(1<<ADNS_SCLK)|(1<<ADNS_NCS);
	PORTB=(1<<ADNS_NCS);

	//Reset the ADNS
	_delay_ms(10);
	adnsWriteByte(ADNSWR|0x3a);
	adnsWriteByte(0x5a);
	adnsEnd();
	//Select 500CPI
	adnsWriteByte(ADNSWR|0x0D);
	adnsWriteByte(0);
	adnsEnd();

	while(1) {
		do {
			//Get all data from sensor
			adnsWriteByte(0x63);
			dx=adnsReadByte();
			dy=adnsReadByte();
			squal=adnsReadByte();
			shut=adnsReadByte()<<8;
			shut|=adnsReadByte();
			adnsReadByte();
			sum=adnsReadByte();
			adnsEnd();
			if (squal<3) {
				OCR0A=0; OCR0B=0;
			}
		} while (squal<3); //Loop while we're picked up

	
		//Correct for the fact that even with the sensor, the device has an offset :/
		count++;
		if (count==6) {
	    	count=0;
	    	x+=dy;
		}

		//See if everything beneath us suddenly went dark
		//If so, set isTurning.
		//isTurning will be 1 for a left turn, 2 for a right turn.
		//After doing a 180 degrees turn, isTurning will stay 3 for a
		//while, which means the bot will go straightforward but ignore the
		//line we pass again when we re-enter the circle.
		if ((sum<sumCal) && (!isTurning)) {
			isTurning=1+(sum&1);
		}
		

#define FACT 4
#define MOTOR_MAX 0xC0
		x+=dx;
		if (isTurning==1 || isTurning==2) {
			//Go left or right, dependant on isTurning.
			if (isTurning==2) {
				OCR0A=MOTOR_MAX;
				OCR0B=0;
			} else {
				OCR0B=MOTOR_MAX;
				OCR0A=0;
			}
			
			if (x>2000 || x<-2000) { //Did we do a 180 degrees turn already?
				isTurning=3;
				x=0; y=0;
				//Recalibrate darkness detection thingy; we're probably not
				//on a line now. 
				sumCal=sum-(sum>>1); 
			}
		} else {
			if (isTurning==3) {
				//Keep isTurning valid for a bit longer so we ignore the
				//black line on re-entry
				y+=dy;
				if (y>3000) isTurning=0;
			}
			
			//Try to go straightforward. Yes, this is weird code, I hope you 
			//you understand it :)
			if (x>10) {
				OCR0A=MOTOR_MAX;
				if ((x*FACT)>MOTOR_MAX) OCR0B=0; else OCR0B=MOTOR_MAX-(x*FACT);
			} else if (x<-10) {
				OCR0B=MOTOR_MAX;
				if (((-x)*FACT)>MOTOR_MAX) OCR0A=0; else OCR0A=MOTOR_MAX-((-x)*FACT);
			} else {
				OCR0A=MOTOR_MAX; OCR0B=MOTOR_MAX;
			}
		}
	}
}
